from __future__ import division, absolute_import, print_function
from .base import ImportanceScoreWRef
from .gradient import Gradient
from kipoi.model import KerasModel, TensorFlowModel
from kipoi.data_utils import numpy_collate

import tempfile
import numpy as np


# Other proposal (this object is passed as an argument to compile())
class DeepLift(ImportanceScoreWRef):
    """
    Wrapper around DeepLIFT
    """

    def __init__(self, model, output_layer,
                 task_idx, preact, mxts_mode='rescale_conv_revealcancel_fc',
                 batch_size=32):
        """
        Args:
          model: Kipoi model
          output_layer (int): selected Keras layer with respect to which the scores should be calculated
          task_idx (int): Node/Neuron within the selected layer with respect to which the score should be calculated
          preact: !NOT YET IMPLEMENTED! Use values prior to activation - for now the default is True!
          mxts_mode: Selected score
          batch_size: Batch size for scoring
        """
        from deeplift.conversion import kerasapi_conversion as kc
        from deeplift.layers import NonlinearMxtsMode

        def get_mxts_mode(mode_name):
            # Labels from examples:
            mxts_modes = {'rescale_conv_revealcancel_fc': NonlinearMxtsMode.DeepLIFT_GenomicsDefault,
                          'revealcancel_all_layers': NonlinearMxtsMode.RevealCancel,
                          'rescale_all_layers': NonlinearMxtsMode.Rescale,
                          'grad_times_inp': NonlinearMxtsMode.Gradient,
                          'guided_backprop': NonlinearMxtsMode.GuidedBackprop}
            return mxts_modes[mode_name]

        self.model = model
        if not self.is_compatible(model):
            raise Exception("Model not compatible with DeepLift")

        self.task_idx = task_idx
        self.batch_size = batch_size

        weight_f = tempfile.mktemp()
        arch_f = tempfile.mktemp()
        model.model.save_weights(weight_f)
        with open(arch_f, "w") as ofh:
            ofh.write(model.model.to_json())
        self.deeplift_model = kc.convert_model_from_saved_files(weight_f, json_file=arch_f,
                                                                nonlinear_mxts_mode=get_mxts_mode(mxts_mode))

        # TODO this code may be useful for future when functional models can be handled too
        self.input_layer_idxs = [0]
        self.output_layers_idxs = [-1]
        """
        input_names = self.model._get_feed_input_names()
        self.input_layer_idxs = []
        self.output_layers_idxs = []
        for input_name in input_names:
            input_layer_name = input_name[:-len("_input")] if input_name.endswith("_input") else input_name
            for i, l in enumerate(self.model.model.layers):
                if l.name == input_layer_name:
                    self.input_layer_idxs.append(i)
        """

        self.fwd_predict_fn = None

        # Now try to find the correct layer:
        if not isinstance(output_layer, int):
            raise Exception("output_layer has to be an integer index of the Keras layer in the Keras model.")

        # TODO: DeepLIFT does not guarantee that the layer naming recapitulates the Keras layer order.
        if output_layer < 0:
            output_layer = len(model.model.layers) + output_layer
        target_layer_idx = [i for i, l in enumerate(self.deeplift_model.get_layers()) if l.name == str(output_layer)][0]

        # Compile the function that computes the contribution scores
        # For sigmoid or softmax outputs, target_layer_idx should be -2 (the default)
        # (See "3.6 Choice of target layer" in https://arxiv.org/abs/1704.02685 for justification)
        # For regression tasks with a linear output, target_layer_idx should be -1
        # (which simply refers to the last layer)
        # If you want the DeepLIFT multipliers instead of the contribution scores, you can use get_target_multipliers_func
        self.deeplift_contribs_func = self.deeplift_model.get_target_contribs_func(
            find_scores_layer_idx=self.input_layer_idxs,
            target_layer_idx=target_layer_idx)

    @classmethod
    def is_compatible(cls, model):
        if model.type != "keras":
            # Support only keras models
            return False
        # Check the Keras backend
        import keras
        import keras.backend as K

        if not int(keras.__version__.split(".")[0]) == 2:
            return False

        # Can only support sequential model since the layer ordering is not 1:1
        if not isinstance(model.model, keras.Sequential):
            return False

        # Backend has to be tensorflow
        if model.backend is None:
            backend = K.backend()
        else:
            backend = model.backend
        return backend == "tensorflow"

    def score(self, input_batch, input_ref):
        """
        Calculate DeepLIFT scores of a given input sequence.
        Args:
          input_batch: Model input data 
        Returns:
          DeepLIFT scores in the same shape / same containers as the input batch.
        """
        x_standardized = self.model._batch_to_list(input_batch)
        ref_standaradized = None
        if input_ref is not None:
            ref_standaradized = self.model._batch_to_list(input_ref)

        scores = self.deeplift_contribs_func(task_idx=self.task_idx,
                                             input_data_list=x_standardized,
                                             input_references_list=ref_standaradized,
                                             batch_size=self.batch_size,
                                             progress_update=1000)

        # TODO DeepLIFT error when using batched execution:
        """
        # run_function_in_batches fails for 
        scores = run_function_in_batches(
            func=self.deeplift_contribs_func,
            input_data_list=x_standardized,
            batch_size=self.batch_size,
            progress_update=1000,
            task_idx=self.task_idx)
        """

        # DeepLIFT returns all samples as a list of individual samples
        scores = [numpy_collate(el) for el in scores]

        # re-format the list-type input back to how the input_batch was:
        scores = self.model._match_to_input(scores, input_batch)
        return scores

    def predict_on_batch(self, input_batch):
        """
        Function that can be used to check the successful model conversion. The output of this function should match 
        the output of the original model when executing .predict(input_batch)
        Args:
          input_batch: Model input data 
        Returns:
          Model predictions
        """
        from deeplift.util import run_function_in_batches
        from deeplift.util import compile_func
        x_standardized = self.model._batch_to_list(input_batch)
        if self.fwd_predict_fn is None:
            # TODO: Once DeepLIFT layer annotation works integrate it here too:
            """
            # identify model output layers:
            self.output_layers_idxs = []
            for output_name in self.model.model.output_names:
                for i, l in enumerate(self.model.model.layers):
                    if l.name == output_name:
                        self.output_layers_idxs.append(i)
            """
            inputs = [self.deeplift_model.get_layers()[i].get_activation_vars() for i in self.input_layer_idxs]
            outputs = [self.deeplift_model.get_layers()[i].get_activation_vars() for i in self.output_layers_idxs]
            self.fwd_predict_fn = compile_func(inputs, outputs)

        preds = run_function_in_batches(
            input_data_list=x_standardized,
            func=self.fwd_predict_fn,
            batch_size=self.batch_size,
            progress_update=None)

        preds = np.array(preds)
        if len(self.output_layers_idxs) == 1:
            preds = preds[0, ...]

        return preds


class IntegratedGradients(ImportanceScoreWRef, Gradient):

    def score(self, input_batch, input_ref):
        grads = super(IntegratedGradients, self).score(input_batch)
        # TODO implement integrated gradients
        # https://github.com/marcoancona/DeepExplain/blob/master/deepexplain/tensorflow/methods.py#L208-L225
        pass


class GradientXInput(ImportanceScoreWRef, Gradient):
    # AbstractGrads implements the

    def score(self, input_batch, input_ref):
        # TODO - handle also the case where input_ref is not a simple array but a list
        return super(GradientXInput, self).score(input_batch) * input_ref


METHODS = {"deeplift": DeepLift,
           "grad*input": GradientXInput,
           "intgrad": IntegratedGradients}



"""CLI
"""
from __future__ import absolute_import
from __future__ import print_function

import argparse
import logging
import os
import sys

from kipoi_interpret.importance_scores.cli import cli_feature_importance

import kipoi
from kipoi.cli.parser_utils import add_model, add_dataloader, file_exists, dir_exists
from kipoi import writers
from kipoi.utils import cd
from kipoi.utils import parse_json_file_str

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def isint(qstr):
    import re
    return bool(re.match("^[0-9]+$", qstr))


# Parse the slice
def parse_filter_slice(in_str):
    if in_str.startswith("(") or in_str.startswith("["):
        in_str_els = in_str.lstrip("([").rstrip(")]").split(",")
        slices = []
        for slice_el in in_str_els:
            slice_el = slice_el.strip(" ")
            if slice_el == "...":
                slices.append(Ellipsis)
            elif isint(slice_el):
                slices.append(int(slice_el))
                if len(in_str_els) == 1:
                    return int(slice_el)
            else:
                # taken from https://stackoverflow.com/questions/680826/python-create-slice-object-from-string
                slices.append(slice(*map(lambda x: int(x.strip()) if x.strip() else None, slice_el.split(':'))))
        return tuple(slices)
    elif isint(in_str):
        return int(in_str)
    else:
        raise Exception("Filter index slice not valid. Allowed values are e.g.: '1', [1:3,...], [:, 0:4, ...]")


def cli_grad(command, raw_args):
    """CLI interface to predict
    """
    # from .main import prepare_batch
    from kipoi.model import GradientMixin
    assert command == "grad"
    from tqdm import tqdm
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='Save gradients and inputs to a hdf5 file.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    parser.add_argument("-i", "--install_req", action='store_true',
                        help="Install required packages from requirements.txt")
    parser.add_argument("-l", "--layer", default=None,
                        help="Which output layer to use to make the predictions. If specified," +
                             "`model.predict_activation_on_batch` will be invoked instead of `model.predict_on_batch`",
                        required=False)
    parser.add_argument("--final_layer",
                        help="Alternatively to `--layer` this flag can be used to indicate that the last layer should "
                             "be used.", action='store_true')
    parser.add_argument("--pre_nonlinearity",
                        help="Flag indicating that it should checked whether the selected output is post activation "
                             "function. If a non-linear activation function is used attempt to use its input. This "
                             "feature is not available for all models.", action='store_true')
    parser.add_argument("-f", "--filter_idx",
                        help="Filter index that should be inspected with gradients. If not set all filters will " +
                             "be used.", default=None)
    parser.add_argument("-a", "--avg_func",
                        help="Averaging function to be applied across selected filters (`--filter_idx`) in " +
                             "layer `--layer`.", choices=GradientMixin.allowed_functions, default="sum")
    parser.add_argument('--selected_fwd_node', help="If the selected layer has multiple inbound connections in "
                                                    "the graph then those can be selected here with an integer "
                                                    "index. Not necessarily supported by all models.",
                        default=None, type=int)
    parser.add_argument('-o', '--output', required=True, nargs="+",
                        help="Output files. File format is inferred from the file path ending. Available file formats are: " +
                             ", ".join(["." + k for k in writers.FILE_SUFFIX_MAP]))
    args = parser.parse_args(raw_args)

    dataloader_kwargs = parse_json_file_str(args.dataloader_args)

    # setup the files
    if not isinstance(args.output, list):
        args.output = [args.output]
    for o in args.output:
        ending = o.split('.')[-1]
        if ending not in writers.FILE_SUFFIX_MAP:
            logger.error("File ending: {0} for file {1} not from {2}".
                         format(ending, o, writers.FILE_SUFFIX_MAP))
            sys.exit(1)
        dir_exists(os.path.dirname(o), logger)
    # --------------------------------------------
    # install args
    if args.install_req:
        kipoi.pipeline.install_model_requirements(args.model,
                                                  args.source,
                                                  and_dataloaders=True)

    layer = args.layer
    if layer is None and not args.final_layer:
        raise Exception("A layer has to be selected explicitely using `--layer` or implicitely by using the"
                        "`--final_layer` flag.")

    # Not a good idea
    # if layer is not None and isint(layer):
    #    logger.warn("Interpreting `--layer` value as integer layer index!")
    #    layer = int(args.layer)

    # load model & dataloader
    model = kipoi.get_model(args.model, args.source)

    if not isinstance(model, GradientMixin):
        raise Exception("Model does not support gradient calculation.")

    if args.dataloader is not None:
        Dl = kipoi.get_dataloader_factory(args.dataloader, args.dataloader_source)
    else:
        Dl = model.default_dataloader

    dataloader_kwargs = kipoi.pipeline.validate_kwargs(Dl, dataloader_kwargs)
    dl = Dl(**dataloader_kwargs)

    filter_idx_parsed = None
    if args.filter_idx is not None:
        filter_idx_parsed = parse_filter_slice(args.filter_idx)

    # setup batching
    it = dl.batch_iter(batch_size=args.batch_size,
                       num_workers=args.num_workers)

    # Setup the writers
    use_writers = []
    for output in args.output:
        ending = output.split('.')[-1]
        W = writers.FILE_SUFFIX_MAP[ending]
        logger.info("Using {0} for file {1}".format(W.__name__, output))
        if ending == "tsv":
            assert W == writers.TsvBatchWriter
            use_writers.append(writers.TsvBatchWriter(file_path=output, nested_sep="/"))
        elif ending == "bed":
            raise Exception("Please use tsv or hdf5 output format.")
        elif ending in ["hdf5", "h5"]:
            assert W == writers.HDF5BatchWriter
            use_writers.append(writers.HDF5BatchWriter(file_path=output))
        else:
            logger.error("Unknown file format: {0}".format(ending))
            sys.exit(1)

    # Loop through the data, make predictions, save the output
    for i, batch in enumerate(tqdm(it)):
        # validate the data schema in the first iteration
        if i == 0 and not Dl.output_schema.compatible_with_batch(batch):
            logger.warn("First batch of data is not compatible with the dataloader schema.")

        # make the prediction
        pred_batch = model.input_grad(batch['inputs'], filter_idx=filter_idx_parsed,
                                      avg_func=args.avg_func, layer=layer, final_layer=args.final_layer,
                                      selected_fwd_node=args.selected_fwd_node,
                                      pre_nonlinearity=args.pre_nonlinearity)

        # write out the predictions, metadata (, inputs, targets)
        # always keep the inputs so that input*grad can be generated!
        # output_batch = prepare_batch(batch, pred_batch, keep_inputs=True)
        output_batch = batch
        output_batch["grads"] = pred_batch
        for writer in use_writers:
            writer.batch_write(output_batch)

    for writer in use_writers:
        writer.close()
    logger.info('Done! Gradients stored in {0}'.format(",".join(args.output)))


def cli_gr_inp_to_file(command, raw_args):
    """ CLI to save seq inputs of grad*input to a bigwig file
    """
    assert command == "gr_inp_to_file"
    parser = argparse.ArgumentParser('kipoi interpret {}'.format(command),
                                     description='Save grad*input in a file.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    parser.add_argument('-f', '--input_file', required=False,
                        help="Input HDF5 file produced from `grad`")
    parser.add_argument('-o', '--output', required=False,
                        help="Output bigwig for bedgraph file")
    parser.add_argument('--sample', required=False, type=int, default=None,
                        help="Input line for which the BigWig file should be generated. If not defined all"
                             "samples will be written.")
    parser.add_argument('--model_input', required=False, default=None,
                        help="Model input name to be used for plotting. " +
                        "As defined in model.yaml. Can be omitted if" +
                        "model only has one input.")
    args = parser.parse_args(raw_args)

    # Check that all the folders exist
    dir_exists(os.path.dirname(args.output), logger)
    # --------------------------------------------
    # install args
    from kipoi_interpret.gradviz import GradPlotter
    from kipoi.writers import BedGraphWriter

    logger.info('Loading gradient results file and model info...')

    gp = GradPlotter.from_hdf5(args.input_file, model=args.model, source=args.source)

    if args.sample is not None:
        samples = [args.sample]
    else:
        samples = list(range(gp.get_num_samples(args.model_input)))

    if args.output.endswith(".bed") or args.output.endswith(".bedgraph"):
        of_obj = BedGraphWriter(args.output)
    else:
        raise Exception("Output file format not supported!")

    logger.info('Writing...')

    for sample in samples:
        gp.write(sample, model_input=args.model_input, writer_obj=of_obj)

    logger.info('Saving...')

    of_obj.close()

    logger.info('Successfully wrote grad*input to file.')


# --------------------------------------------
# CLI commands


command_functions = {
    'feature_importance': cli_feature_importance,

    # Deprecate
    # 'gr_inp_to_file': cli_gr_inp_to_file,  # TODO - rename this to grad_input?
    # 'ism': cli_create_mutation_map,
}
commands_str = ', '.join(command_functions.keys())

# TODO - shall we call the grad commands: saliency?
parser = argparse.ArgumentParser(
    description='Kipoi model-zoo command line tool',
    usage='''kipoi interepret <command> [-h] ...

    # Available sub-commands:
    feature_importance    Compute the feature importance
    ''')
parser.add_argument('command', help='Subcommand to run; possible commands: {}'.format(commands_str))

CLI_DESCRIPTION = "Compute feature importance scores"


def cli_main(command, raw_args):
    args = parser.parse_args(raw_args[0:1])
    if args.command not in command_functions:
        parser.print_help()
        parser.exit(
            status=1,
            message='\nCommand `{}` not found. Possible commands: {}\n'.format(
                args.command, commands_str))
    command_fn = command_functions[args.command]
    command_fn(args.command, raw_args[1:])


if __name__ == '__main__':
    command = sys.argv[1]
    raw_args = sys.argv[1:]
    cli_main(command, raw_args)



from __future__ import division, absolute_import, print_function
from tqdm import tqdm
from kipoi.utils import merge_dicts
from kipoi.data_utils import numpy_collate_concat
import abc
# Abstract declaration of different interpretation APIs, plus any
# common functionality


# Abstract importance score
class ImportanceScore(object):

    @classmethod
    def is_compatible(self, model):
        """Test if the model is compatible with the API
        """
        raise NotImplementedError()

    def score(self, input_batch):
        """Score a particular example
        """
        raise NotImplementedError()


# Importance score that requires also the reference
class ImportanceScoreWRef(ImportanceScore):

    @abc.abstractmethod
    def score(self, input_batch, input_ref):
        # read the data loaders into arrays as needed
        # instantiate the reference generator using the kwargs
        # call score_func accordingly
        # TODO: implement
        raise NotImplementedError()


# --------------------------------------------

def available_methods():
    """Get all available methods
    """
    from . import ism, gradient, referencebased
    int_modules = [ism, gradient, referencebased]

    available_methods = {}
    for m in int_modules:
        available_methods = merge_dicts(available_methods, m.METHODS)
    return available_methods


def get_importance_score(importance_score):
    """Get the importance score
    """
    if isinstance(importance_score, ImportanceScore):
        return importance_score
    methods = available_methods()
    if importance_score not in methods:
        raise ValueError("Importance score {0} not found. Valid scores are: {1}".
                         format(importance_score, methods))
    return methods[importance_score]


# --------------------------------------------


def feature_importance(model,
                       dataloader,
                       importance_score,
                       importance_score_kwargs={},
                       batch_size=32,
                       num_workers=0):
    """Return feature importance scores
    """
    ImpScore = get_importance_score(importance_score)
    if not ImpScore.is_compatible(model):
        raise ValueError("model not compatible with score: {0}".format(importance_score))
    impscore = ImpScore(model, **importance_score_kwargs)

    def append_key(d, k, v):
        d[k] = v
        return d

    # TODO - handle the reference-based importance scores...
    return numpy_collate_concat([append_key(batch, "importance_scores", impscore.score(batch['inputs']))
                                 for batch in tqdm(dataloader.batch_iter(batch_size=batch_size,
                                                                         num_workers=num_workers))])



"""CLI
"""
from __future__ import absolute_import
from __future__ import print_function

import argparse
import logging
import os
import sys

#from kipoi_interpret.feature_importance import available_importance_scores, get_importance_score

from tqdm import tqdm
import kipoi
from kipoi.cli.parser_utils import add_model, add_dataloader, file_exists, dir_exists
from kipoi import writers
from kipoi.utils import parse_json_file_str

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def cli_feature_importance(command, raw_args):
    """CLI interface to predict
    """
    # from .main import prepare_batch
    assert command == "feature_importance"
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='Save gradients and inputs to a hdf5 file.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    parser.add_argument("--imp_score", help="Importance score name", choices=available_importance_scores())
    parser.add_argument("--imp_score_kwargs", help="Importance score kwargs")
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    # TODO - handle the reference-based importance scores...

    # io
    parser.add_argument('-o', '--output', required=True, nargs="+",
                        help="Output files. File format is inferred from the file path ending. Available file formats are: " +
                             ", ".join(["." + k for k in writers.FILE_SUFFIX_MAP]))
    args = parser.parse_args(raw_args)

    dataloader_kwargs = parse_json_file_str(args.dataloader_args)
    imp_score_kwargs = parse_json_file_str(args.imp_score_kwargs)

    # setup the files
    if not isinstance(args.output, list):
        args.output = [args.output]
    for o in args.output:
        ending = o.split('.')[-1]
        if ending not in writers.FILE_SUFFIX_MAP:
            logger.error("File ending: {0} for file {1} not from {2}".
                         format(ending, o, writers.FILE_SUFFIX_MAP))
            sys.exit(1)
        dir_exists(os.path.dirname(o), logger)
    # --------------------------------------------
    # install args
    if args.install_req:
        kipoi.pipeline.install_model_requirements(args.model,
                                                  args.source,
                                                  and_dataloaders=True)

    # load model & dataloader
    model = kipoi.get_model(args.model, args.source, with_dataloader=args.dataloader is None)

    if args.dataloader is not None:
        Dl = kipoi.get_dataloader_factory(args.dataloader, args.dataloader_source)
    else:
        Dl = model.default_dataloader

    dataloader_kwargs = kipoi.pipeline.validate_kwargs(Dl, dataloader_kwargs)
    dl = Dl(**dataloader_kwargs)

    # get_importance_score
    ImpScore = get_importance_score(args.imp_score)
    if not ImpScore.is_compatible(model):
        raise ValueError("model not compatible with score: {0}".format(args.imp_score))
    impscore = ImpScore(model, **imp_score_kwargs)

    # setup batching
    it = dl.batch_iter(batch_size=args.batch_size,
                       num_workers=args.num_workers)

    # Setup the writers
    use_writers = []
    for output in args.output:
        ending = output.split('.')[-1]
        W = writers.FILE_SUFFIX_MAP[ending]
        logger.info("Using {0} for file {1}".format(W.__name__, output))
        if ending == "tsv":
            assert W == writers.TsvBatchWriter
            use_writers.append(writers.TsvBatchWriter(file_path=output, nested_sep="/"))
        elif ending == "bed":
            raise Exception("Please use tsv or hdf5 output format.")
        elif ending in ["hdf5", "h5"]:
            assert W == writers.HDF5BatchWriter
            use_writers.append(writers.HDF5BatchWriter(file_path=output))
        else:
            logger.error("Unknown file format: {0}".format(ending))
            sys.exit(1)

    # Loop through the data, make predictions, save the output
    for i, batch in enumerate(tqdm(it)):
        # validate the data schema in the first iteration
        if i == 0 and not Dl.output_schema.compatible_with_batch(batch):
            logger.warn("First batch of data is not compatible with the dataloader schema.")

        # make the prediction
        # TODO - handle the reference-based importance scores...
        importance_scores = impscore.score(batch['inputs'])

        # write out the predictions, metadata (, inputs, targets)
        # always keep the inputs so that input*grad can be generated!
        # output_batch = prepare_batch(batch, pred_batch, keep_inputs=True)
        output_batch = batch
        output_batch["importance_scores"] = importance_scores
        for writer in use_writers:
            writer.batch_write(output_batch)

    for writer in use_writers:
        writer.close()
    logger.info('Done! Importance scores stored in {0}'.format(",".join(args.output)))



import numpy as np
import collections
from collections import OrderedDict

def get_model_input_special_type(model, input_id):
    inputs = model.schema.inputs
    if isinstance(inputs, collections.Mapping) or isinstance(inputs, collections.Sequence):
        return inputs[input_id].special_type
    else:
        return inputs.special_type

def get_model_input(batch, input_id=None):
    """
    Get model input from batch
    batch: batch of model input samples
    """
    if isinstance(batch, dict) or isinstance(batch, list):
        assert input_id is not None
        return batch[input_id]
    else:
        return batch



def set_model_input(batch, value, input_id=None):
    if isinstance(batch, dict) or isinstance(batch, list):
        assert input_id is not None
        batch[input_id] = value
        return batch
    else:
        return value



def apply_within(data1, data2, function, **kwargs):
    if not type(data1) == type(data2):
        raise Exception("data1 and data2 have to be of the same type!")
    if isinstance(data1, dict):
        if not set(data1.keys()) == set(data2.keys()):
            raise Exception("data1 and data2 dictionaries have to have the same keys!")
        out = {}
        if isinstance(data1, OrderedDict):
            out = OrderedDict()
        for k in data1:
            out[k] = apply_within(data1[k], data2[k], function, **kwargs)
    elif isinstance(data1, list):
        if not len(data1) == len(data2):
            raise Exception("data1 and data2 lists have to have the same length!")
        out = []
        for el1, el2 in zip(data1, data2):
            out.append(apply_within(el1, el2, function, **kwargs))
    else:
        out = function(data1, data2, **kwargs)
    return out



from kipoi import readers
# from kipoi.postprocessing.variant_effects import ModelInfoExtractor
# from kipoi.postprocessing.variant_effects.utils.generic import _get_metadata_name, _get_seq_shape
# from kipoi.postprocessing.variant_effects.utils import ReshapeDna
import kipoi
import os
import numpy as np
import logging
from kipoi.utils import cd

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def grad_x_input(input, grad):
    return input * grad


def get_selector(dim, slice_at_dim):
    """
    Place `slice_at_dim` at dimension `dim` for setting or setting items
    """
    if dim >= 0:
        selector = [slice(None) for i in range(dim)] + [slice_at_dim] + [Ellipsis]
    else:
        selector = [Ellipsis] + [slice_at_dim] + [slice(None) for i in range((dim) * (-1) - 1)]
    selector = tuple(selector)
    return selector


# TODO - why is this called gradPlotter? In theory, this could be any other importance score...

# - can't depend on anything from kipoi.postprocessing..

# class GradPlotter(object):
#     """
#     Class for plotting gradients. Results can be loaded from a HDF5 file or directly from returns of
#     model.input_grad(...)
#     """

#     def __init__(self, data, model, source="kipoi", grad_preds=None):
#         """
#         Arguments:
#             data: model input data batch
#             model: model name as used for running `model.input_grad(...)`
#             source: model source as used for running `model.input_grad(...)`
#             grad_preds: return value of `model.input_grad(...)`. Can alternatively already be present in `data`
#             argument under the key `preds`. In that case `grad_preds` may be None.
#         """
#         # TODO - prettify this init
#         self.data = data
#         if grad_preds is not None:
#             self.data['grads'] = grad_preds
#         else:
#             assert 'grads' in self.data

#         # TODO: R: Instead of copying from kipoi.model should we rather have a get_model_descr
#         #             -- Z: this function is indeed available
#         # TODO-cont: funcion that is also called from get_model
#         # Taken from get_model
#         source_name = source
#         source = kipoi.config.get_source(source)
#         md = source.get_model_descr(model)

#         if ":" in md.default_dataloader:
#             dl_source, dl_path = md.default_dataloader.split(":")
#         else:
#             dl_source = source_name
#             dl_path = md.default_dataloader

#         # allow to use relative and absolute paths for referring to the dataloader
#         default_dataloader_path = os.path.join("/" + model, dl_path)[1:]
#         # This one loads the model!!
#         # default_dataloader = kipoi.get_dataloader_factory(default_dataloader_path,
#         #                                                  dl_source)

#         # TODO: Is there a nicer way of getting ahold of the dataloader description?
#         yaml_path = source.pull_dataloader(default_dataloader_path)
#         dataloader_dir = os.path.dirname(yaml_path)
#         from kipoi.components import DataLoaderDescription
#         with cd(dataloader_dir):
#             dl = DataLoaderDescription.load(os.path.basename(yaml_path))
#             default_dataloader = dl

#         try:
#             self.mie = ModelInfoExtractor(md, default_dataloader)
#         except:
#             logger.warn("Model is not enabled for variant effect prediction hence it is unclear whether there is a DNA "
#                         "sequence input, so (automatic) seqlogo plots are not available for this model.")
#             self.mie = None
#         self.md = md
#         self.dataloader = default_dataloader

#         # how can the correct model input be selected
#         self.get_dataset, self.model_input_keylist = self._get_ds_extractor(md.schema.inputs)

#     @classmethod
#     def from_hdf5(self, results_fname, model, source="kipoi"):
#         """
#         Arguments:
#             results_fname: HDF5 file produced by running `model.input_grad(...)`. The file has to contain
#             also the model input!
#             model: model name as used for running `model.input_grad(...)`
#             source: model source as used for running `model.input_grad(...)`
#         """
#         data = readers.HDF5Reader.load(results_fname, unflatten=True)
#         return self(data, model=model, source=source)

#     def _get_ds_extractor(self, model_input_schema):
#         from kipoi.components import ArraySchema
#         # return data selection function + labels
#         if isinstance(model_input_schema, ArraySchema):
#             return self._select_ds_ndarray, [model_input_schema.name]
#         elif isinstance(model_input_schema, list):
#             return self._select_ds_list, [el.name for el in model_input_schema]
#         elif isinstance(model_input_schema, dict):
#             return self._select_ds_dict, [el for el in model_input_schema]

#     def _select_ds_ndarray(self, index, dataset):
#         if index is not None:
#             assert (index == 0) or (self.model_input_keylist[0] == index)
#         return dataset

#     def _select_ds_list(self, index, dataset):
#         if isinstance(index, int):
#             return dataset[index]
#         else:
#             int_ind = self.model_input_keylist.index(index)
#             return dataset[int_ind]

#     def _select_ds_dict(self, index, dataset):
#         return dataset[index]

#     def _verify_model_input(self, model_input):
#         if model_input is None:
#             if len(self.model_input_keylist) != 1:
#                 raise Exception("model_input cannot be None for models that have multiple "
#                                 "inputs: %s" % str(self.model_input_keylist))
#             model_input = self.model_input_keylist[0]
#         return model_input

#     def get_num_samples(self, model_input=None):
#         """
#         Get number of samples present in the dataset (size of 0th dimension of the model input).

#         Arguments:
#             model_input: Number of samples will be returned for that model input.
#         """
#         model_input = self._verify_model_input(model_input)
#         input = self.get_dataset(model_input, self.data["inputs"])
#         return input.shape[0]

#     @staticmethod
#     def _infer_seq_dim(len, array_shape):
#         sel = np.array(array_shape) == len
#         if sel.sum() == 1:
#             return np.where(sel)[0][0]
#         return None

#     def _preprocess_values(self, sample, model_input=None,
#                            limit_region=None, limit_region_genomic=None,
#                            transform_fn=grad_x_input,
#                            seq_dim=None,
#                            requires_region_info=False,
#                            requires_seq_dim=False):
#         def raise_missing_metadata():
#             raise Exception("limit_region_genomic can only be used with a `model_input` that has an associated"
#                             "metadata field with at least the following entries: 'chrom', 'start', 'end' or type"
#                             "GenomicRanges.")

#         def get_metadata_cse():
#             metadata_field = None
#             try:
#                 metadata_field = _get_metadata_name(self.dataloader, model_input)
#             except ValueError:
#                 raise_missing_metadata()
#             mf = self.data['metadata'][metadata_field]
#             if not all([el in mf for el in ["chr", "start", "end"]]):
#                 raise_missing_metadata()
#             cse = {k: np.squeeze(mf[k]) for k in ["chr", "start", "end"]}
#             for k in ["chr", "start", "end"]:
#                 cse[k] = np.squeeze(mf[k])
#                 # if only one sample then squeeze does too much, so correct that:
#                 if len(cse[k].shape) == 0:
#                     cse[k] = np.array([cse[k]])
#                 if len(cse[k].shape) != 1:
#                     raise Exception("Invalid metadata format for field ['%s']['%s'] with shape: %s" %
#                                     (metadata_field, k, str(mf[k].shape)))
#             return cse

#         inputs = self.data["inputs"]
#         gradients = self.data["grads"]

#         model_input = self._verify_model_input(model_input)

#         input = self.get_dataset(model_input, inputs)
#         gradient = self.get_dataset(model_input, gradients)

#         is_onehot_seq = False
#         if (self.mie is not None) and (model_input in self.mie.get_mutatable_inputs(only_one_hot=True)):
#             is_onehot_seq = True

#         if requires_region_info:
#             mf = get_metadata_cse()
#             mr_chr = mf["chr"][sample]
#             mr_start = mf["start"][sample]
#             mr_end = mf["end"][sample]
#         else:
#             mr_chr = None
#             mr_start = None
#             mr_end = None

#         if (limit_region_genomic is not None):
#             if not isinstance(limit_region_genomic, tuple):
#                 raise Exception("`limit_region_genomic` has to be a tuple of (start, end) genomic coordinates!")
#             mf = get_metadata_cse()
#             mr_chr = mf["chr"][sample]
#             mr_start = mf["start"][sample]
#             mr_end = mf["end"][sample]
#             if any([(el < mr_start) or (el > mr_end) for el in list(limit_region_genomic)]):
#                 raise Exception("`limit_region_genomic` has to lie within: %s" % str([mr_start, mr_end]))
#             limit_region = (limit_region_genomic[0] - mr_start, limit_region_genomic[1] - mr_start,)
#             mr_start, mr_end = limit_region_genomic

#         elif (limit_region is not None):
#             if not isinstance(limit_region, tuple):
#                 raise Exception("`limit_region` has to be a tuple of (start, end) array indices!")
#             if mr_start is not None:
#                 mr_end = mr_start + limit_region_genomic[1] + 1
#                 mr_start = mr_start + limit_region_genomic[0]

#         if is_onehot_seq:
#             # convert to standard layout
#             dna_reshaper = ReshapeDna(_get_seq_shape(self.dataloader, model_input))
#             input_reshaped = dna_reshaper.to_standard(input)[sample, ...]
#             gradient_reshaped = dna_reshaper.to_standard(gradient)[sample, ...]
#             seq_dim = 0

#             values = transform_fn(input_reshaped, gradient_reshaped)

#             if limit_region is not None:
#                 values = values[limit_region[0]:limit_region[1], :]

#             return values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim

#         else:
#             if requires_seq_dim or (limit_region is not None):
#                 if seq_dim is None:
#                     mf = get_metadata_cse()
#                     seq_len = mf["end"][sample] - mf["start"][sample]
#                     seq_dim = self._infer_seq_dim(seq_len, input.shape[1:])
#                     if seq_dim is None:
#                         raise Exception("seq_dim was not defined and could not be inferred for array dimensions %s and"
#                                         "sequence length %d. seq_dim is required if `rc_plot` or "
#                                         "`limit_region_genomic` or `seq_plotter_obj` on a non-DNAsequence input "
#                                         "are used!")

#             values = transform_fn(input[sample, ...], gradient[sample, ...])
#             if limit_region is not None:
#                 lr_sel = get_selector(seq_dim, slice(limit_region[0], limit_region[1]))
#                 values = values.__getitem__(lr_sel)

#             return values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim

#     def plot(self, sample, model_input=None, ax=None,
#              limit_region=None, limit_region_genomic=None, rc_plot=False,
#              transform_fn=grad_x_input, seq_dim=None, additional_plot_fns=None, **heatmap_kwargs):
#         """
#         Plot grad*input for one sample in the data (batch). If the selected model input is tagged as "DNASequence" and
#         the model variant effect prediction activated (here only necessary for parsing model info), then values
#         returned by `transform_fn` are displayed as Seqlogo plots. The default transform_fn is grad*input.
#         If the selected model input is not tagged as "DNASequence" or the model is not activated for variant effect
#         prediction then a heatmap will be generated with additional `heatmap_kwargs` if given.
#         If a heatmap should be produced then the input can only be the batch sample axis + 2D, otherwise the heatmap
#         fails.

#         Arguments:
#             sample: Sample in the batch (integer)
#             model_input: Name of the model input that should be plotted (can be omitted for models with only one
#             input)
#             ax: axis object to be passed to the plotting functions.
#             limit_region: Tuple. Limits the plot to a subset in the sequence dimension (seq_dim)
#             limit_region_genomic: Tuple. Like `limit_region`, but genomic coordinates - no chromosome
#             rc_plot: Reverse-complement the plot. If model_input is not "DNASequence" then only reverse
#             transform_fn: Function fn(input, grad). Default is input*grad
#             seq_dim: Dimension of the sequence. Used for reversing the order in `rc_plot`. Can be omitted for
#             "DNASequence" model inputs. If not given, but needed it will be attempted to be inferred from array
#             dimensions and metadata sequence length.
#             additional_plot_fns: List of functions fn(chrom, start, end, ax) that will be executed after the main
#             plotting routine
#         """

#         requires_region_info = additional_plot_fns is not None

#         values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim = self._preprocess_values(sample,
#                                                                                            model_input=model_input,
#                                                                                            limit_region=limit_region,
#                                                                                            limit_region_genomic=limit_region_genomic,
#                                                                                            transform_fn=transform_fn,
#                                                                                            seq_dim=seq_dim,
#                                                                                            requires_region_info=requires_region_info,
#                                                                                            requires_seq_dim=True)

#         if ax is None:
#             import matplotlib.pyplot as plt
#             plt.figure(figsize=(20, 4))
#             ax = plt.subplot(1, 1, 1)

#         import seaborn as sns
#         if is_onehot_seq:
#             # Reverse-complement only for the plot itself
#             if rc_plot:
#                 values = values[::-1, ::-1]

#             from kipoi.external.concise.seqplotting_deps import seqlogo
#             seqlogo(values, ax=ax)
#             ax.axes.get_xaxis().set_visible(False)
#             sns.despine(trim=True, bottom=True, ax=ax)

#         else:
#             if rc_plot:
#                 rc_sel = get_selector(seq_dim, slice(None, None, -1))
#                 values = values.__getitem__(rc_sel)

#             sns.heatmap(values, ax=ax, **heatmap_kwargs)

#         if (additional_plot_fns is not None):
#             for plot_fn in additional_plot_fns:
#                 plot_fn(chrom=mr_chr, start=mr_start, end=mr_end, ax=ax)

#     def write(self, sample, writer_obj, model_input=None, limit_region=None, limit_region_genomic=None,
#               transform_fn=grad_x_input, seq_dim=None):
#         """
#         Write grad*input for one sample in the data (batch). If the selected model input is tagged as "DNASequence" and
#         the model variant effect prediction activated (here only necessary for parsing model info), then values
#         returned by `transform_fn` are stored as data in the `writer_obj`. The default transform_fn is grad*input.
#         If the selected model input is not tagged as "DNASequence" or the model is not activated for variant effect
#         prediction then higher-dimensional data will be compressed into a 1D representation by summing over all axes
#         except for the one defined in `seq_dim`.

#         Arguments:
#             sample: Sample in the batch (integer)
#             writer_obj: Is a instance of a subclass of "RegionWriter". "region_write()" will be called once per
#             call of this function.
#             model_input: Name of the model input that should be written (can be omitted for models with only one
#             input)
#             limit_region: Tuple. Limits the values to a subset in the sequence dimension (seq_dim)
#             limit_region_genomic: Tuple. Like `limit_region`, but genomic coordinates - no chromosome
#             transform_fn: Function fn(input, grad). Default is input*grad
#             seq_dim: Dimension of the sequence. Used for reversing the order in `rc_plot`. Can be omitted for
#             "DNASequence" model inputs. If not given, but needed it will be attempted to be inferred from array
#             dimensions and metadata sequence length.
#         """
#         values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim = self._preprocess_values(sample,
#                                                                                            model_input=model_input,
#                                                                                            limit_region=limit_region,
#                                                                                            limit_region_genomic=limit_region_genomic,
#                                                                                            transform_fn=transform_fn,
#                                                                                            seq_dim=seq_dim,
#                                                                                            requires_region_info=True,
#                                                                                            requires_seq_dim=True)

#         regions = {"chr": [mr_chr], "start": [mr_start], "end": [mr_end]}

#         if is_onehot_seq:
#             writer_obj.region_write(regions, values)

#         else:
#             # now compress them down by summation so that only the seq_dim is left
#             values_summed = values
#             for dim in range(0, len(values.shape)):
#                 if dim != seq_dim:
#                     values_summed = np.sum(values_summed, axis=dim, keepdims=True)
#             values_summed = np.squeeze(values_summed)
#             writer_obj.region_write(regions, values_summed)



from __future__ import absolute_import

__author__ = 'Kipoi team'
__email__ = 'avsec@in.tum.de'
__version__ = '0.1.0'

# from .vis import GradPlotter

# Required by kipoi
from .cli import cli_main
# kipoi-interpret doesn' need any global variables
DataloaderParser = None
ModelParser = None



from __future__ import division, absolute_import, print_function
from .base import ImportanceScore
import numpy as np


class Gradient(ImportanceScore):

    def __init__(self, model,
                 filter_idx=None,
                 avg_func="sum",
                 layer=None,
                 selected_fwd_node=None,
                 pre_nonlinearity=False):
        """
        Args:
          model: Kipoi model
          layer": Which output layer to use to make the predictions. If not specified, the final layer will be used.
          pre_nonlinearity: boolean flag indicating that it should checked whether the selected output is post activation
                   function. If a non-linear activation function is used attempt to use its input.
          filter_idx: Filter index that should be inspected with gradients. If not set all filters will be used.
          avg_func: Averaging function to be applied across selected filters (`--filter_idx`) in layer `--layer`."
          selected_fwd_node: If the selected layer has multiple inbound connections in
             the graph then those can be selected here with an integer
             index. Not necessarily supported by all models.
        """
        self.model = model

        self.filter_idx = filter_idx
        self.avg_func = avg_func
        self.layer = layer
        self.selected_fwd_node = selected_fwd_node
        self.pre_nonlinearity = pre_nonlinearity

    @classmethod
    def is_compatible(self, model):
        """Requires the gradient method to be implemented

        Args:
          model: Model instance or ModelDescription
        """
        if hasattr(model, 'input_grad'):
            return True
        if model.type in ["keras", "pytorch", "tensorflow"]:
            return True
        return False

    def score(self, input_batch):
        """
        Calculate gradients of a given input sequence.
        Args:
          input_batch: Model input data 
        Returns:
          Gradients in the same shape / same containers as the input batch.
        """
        return self.model.input_grad(input_batch,
                                     filter_idx=self.filter_idx,
                                     avg_func=self.avg_func,
                                     layer=self.layer,
                                     final_layer=self.layer is None,
                                     selected_fwd_node=self.selected_fwd_node,
                                     pre_nonlinearity=self.pre_nonlinearity)


class Saliency(Gradient):

    def score(self, input_batch):
        # TODO - apply the abs function to all list/dict elements if necessary
        return np.abs(super(Saliency, self).score(self, input_batch))


METHODS = {"saliency": Saliency,
           "grad": Gradient}



from __future__ import division, absolute_import, print_function
from .base import ImportanceScore
import numpy as np
from kipoi.data_utils import numpy_collate_concat, numpy_collate
from kipoi_interpret.utils import get_model_input, set_model_input
from kipoi.data_utils import get_dataset_item
from kipoi_interpret.utils import apply_within
from .ism_scores import Score
import copy


def parse_score_str(score, **init_kwargs):
    if isinstance(score, Score):
        return score
    else:
        return Score.from_string(score)(**init_kwargs)


def parse_scores(scores, score_kwargs):
    if score_kwargs is not None:
        if not len(scores) == len(score_kwargs):
            raise Exception("If score_kwargs is not none then at least an empty dictionary has to be passed for "
                            "every entry in `scores`.")
        return [parse_score_str(score, **kwargs) for score, kwargs in zip(scores, score_kwargs)]
    else:
        return [parse_score_str(score) for score in scores]


class Mutation(ImportanceScore):
    """ISM for working with one-hot encoded inputs. 
    """

    def __init__(self, model, model_input, scores=['diff'], score_kwargs=None, batch_size=32, output_sel_fn = None,
                 category_dim = 1):
        """
        Args:
          model: Kipoi model
          model_input: which model input to mutate
          scores: a list of score names or score instances
          batch_size: batch size for calls to prediction. This is independent from the size of batch
            used with the `score` method.
          score_kwargs: Initialisation keyword arguments for `scores`. If not None then it is a list of kwargs 
            dictionaries of the same length as `scores`
          output_sel_fn: Function used to select a model output. Only the selected output will be reported as a return 
            value.
          category_dim: Dimension in which the the one-hot category is stored. e.g. for a one-hot encoded DNA-sequence
            array with input shape (1000, 4) for a single sample, `category_dim` is 1, for (4, 1000) `category_dim`
            is 0. In the given dimension only one value is allowed to be non-zero, which is the selected one.
        """
        self.model = model
        self.model_input = model_input
        self.batch_size = batch_size
        self.scores = parse_scores(scores, score_kwargs)
        self.output_sel_fn = output_sel_fn
        self.category_dim = category_dim

    def is_compatible(self, model):
        return True

    def mutate_sample(self, onehot_input):
        it = np.ndenumerate(onehot_input)
        for idx, in_val in it:
            if in_val == 1:
                continue
            zero_sel = list(idx)
            zero_sel[self.category_dim] = slice(None)
            output_onehot = onehot_input.copy()
            output_onehot.__setitem__(tuple(zero_sel), 0)
            output_onehot.__setitem__(idx, 1)
            yield output_onehot, idx

    def mutate_sample_batched(self, onehot_input):
        return_samples = []
        return_indexes = []
        for mutated, indexes in self.mutate_sample(onehot_input):
            return_samples.append(mutated)
            return_indexes.append(indexes)
            if len(return_samples) == self.batch_size:
                yield return_samples, return_indexes
                return_samples = []
                return_indexes = []
        if len(return_samples) != 0:
            yield return_samples, return_indexes


    def score(self, input_batch):
        """
        Args:
          input_batch: Input batch that should be scored.
                
        Returns:
          A list of length: len(`scores`). Every element of the list is a stacked list of depth D if the model input
          is D-dimensional with identcal shape. Every entry of that list then contains the scores of the model output 
          selected by `output_sel_fn`. Values are `None` if the input_batch already had a `1` at that position.
        """

        ref = self.model.predict_on_batch(input_batch)
        scores = []
        for sample_i in range(input_batch[self.model_input].shape[0]):
            # get the full set of model inputs for the selected sample
            sample_set = get_dataset_item(input_batch, sample_i)
            # get the reference output for this sample
            ref_sample_pred = get_dataset_item(ref, sample_i)
            # Apply the output selection function if defined
            if self.output_sel_fn is not None:
                ref_sample_pred = self.output_sel_fn(ref_sample_pred)
            # get the one-hot encoded reference input array
            input_sample = get_model_input(sample_set, input_id=self.model_input)
            # where we keep the scores - scores are lists (ordered by diff method) of ndarrays, lists or dictionaries - whatever is returned by the model
            score = np.empty(input_sample.shape, dtype=object)
            score[:] = None
            for alt_batch, alt_idxs in self.mutate_sample_batched(input_sample):
                #
                num_samples = len(alt_batch)
                mult_set = numpy_collate([sample_set]*num_samples)
                mult_set = set_model_input(mult_set, numpy_collate(alt_batch), input_id=self.model_input)
                alt = self.model.predict_on_batch(mult_set)
                for alt_sample_i in range(num_samples):
                    alt_sample = get_dataset_item(alt, alt_sample_i)
                    # Apply the output selection function if defined
                    if self.output_sel_fn is not None:
                        alt_sample = self.output_sel_fn(alt_sample)
                    # Apply scores across all model outputs for ref and alt
                    output_scores = [apply_within(ref_sample_pred, alt_sample, scr) for scr in self.scores]
                    score.__setitem__(alt_idxs[alt_sample_i], output_scores)
            scores.append(score.tolist())

        return scores

    # def score_from_cli(self, input_onehot_fasta_data_loader_config,
    #                    io_batch_size=100,
    #                    **other_kwargs):
    #     # read the data to create input_onehot_fasta
    #     # call score_func accordingly
    #     assert False

# 'zero': (DummyZero, 0),
# 'saliency': (Saliency, 1),
# 'grad*input': (GradientXInput, 2),
# 'intgrad': (IntegratedGradients, 3),
# 'elrp': (EpsilonLRP, 4),
# 'deeplift': (DeepLIFTRescale, 5),
# 'occlusion': (Occlusion, 6)


class Occlusion(ImportanceScore):

    def __init__(self, model, model_input, scores=['diff']):
        """
        Args:
          model: Kipoi model
          model_input: which model input to mutate
          scores: a list of score names or score instances
        """
        self.model = model
        self.model_input = model_input
        self.scores = parse_scores(scores)

    def is_compatible(self, model):
        # ISM is compatible with all the models
        return True

    def score(self, input_batch):
        # TODO - loop through all the possible elements of the array and set them to zero
        def get_mask(idx, shape):
            ones = np.ones(shape)
            ones[idx] = 0
            return ones

        input_batch[self.model_input]
        pass


METHODS = {"occlusion": Occlusion,
           "mutation": Mutation}
